/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2025 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.20.0.1                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------

Purpose : BSP for embOS Linux simulation.

Additional information:
  This BSP.c prints the state of the LEDs after setting, clearing
  or toggling to stdout to visualize the execution of the simulation
  when used with the OS_StartLEDBlink.c sample application.
  Writing to stdout should be disabled by setting BSP_ENABLE_PRINT_LEDS
  to 0 or by rewriting or not using this file, if your application
  already uses e.g. printf() to write to the stdout.
*/

#include "RTOS.h"
#include "BSP.h"
#include <stdio.h>
#include <stdlib.h>

/*********************************************************************
*
*       Configuration
*
**********************************************************************
*/

//
// If BSP_ENABLE_PRINT_LEDS is 1 the state of the LEDs is printed to stdout after an LED is set, cleared or toggled.
//
#ifndef   BSP_ENABLE_PRINT_LEDS
  #define BSP_ENABLE_PRINT_LEDS  (1)
#endif

//
// If BSP_CLEAR_SCREEN is 1 the screen is cleared before the state of the LEDs is printed.
// If BSP_CLEAR_SCREEN is 0 a newline is printed after all LED states were printed.
//
#ifndef   BSP_CLEAR_SCREEN
  #define BSP_CLEAR_SCREEN  (1)
#endif

//
// Specifies the number of LEDs.
//
#ifndef   BSP_MAX_LEDS
  #define BSP_MAX_LEDS  (4)
#endif

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

static OS_MUTEX _Lock;
static char     _LEDState[BSP_MAX_LEDS];
#if (BSP_ENABLE_PRINT_LEDS != 0)
static char     _Symbol[2] = {' ', 'x'};
#endif

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _PrintLEDState()
*/
#if (BSP_ENABLE_PRINT_LEDS != 0)
static void _PrintLEDState(void) {
  int i;

#if (BSP_CLEAR_SCREEN != 0)
  printf("\033[H\033[J");
#endif
  for (i = 0; i < BSP_MAX_LEDS; ++i) {
    printf("[%c] ", _Symbol[(int)_LEDState[i]]);
  }
#if (BSP_CLEAR_SCREEN == 0)
  printf("\n");
#endif
  fflush(stdout);
}
#endif

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       BSP_Init()
*/
void BSP_Init(void) {
  OS_MUTEX_Create(&_Lock);
}

/*********************************************************************
*
*       BSP_SetLED()
*/
void BSP_SetLED(int Index) {
  if (Index < BSP_MAX_LEDS) {
    OS_MUTEX_LockBlocked(&_Lock);
    _LEDState[Index] = 1;
#if (BSP_ENABLE_PRINT_LEDS != 0)
    _PrintLEDState();
#endif
    OS_MUTEX_Unlock(&_Lock);
  }
}

/*********************************************************************
*
*       BSP_ClrLED()
*/
void BSP_ClrLED(int Index) {
  if (Index < BSP_MAX_LEDS) {
    OS_MUTEX_LockBlocked(&_Lock);
    _LEDState[Index] = 0;
#if (BSP_ENABLE_PRINT_LEDS != 0)
    _PrintLEDState();
#endif
    OS_MUTEX_Unlock(&_Lock);
  }
}

/*********************************************************************
*
*       BSP_ToggleLED()
*/
void BSP_ToggleLED(int Index) {
  if (Index < BSP_MAX_LEDS) {
    OS_MUTEX_LockBlocked(&_Lock);
    _LEDState[Index] ^= 1;
#if (BSP_ENABLE_PRINT_LEDS != 0)
    _PrintLEDState();
#endif
    OS_MUTEX_Unlock(&_Lock);
  }
}

/*************************** End of file ****************************/