/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2025 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.20.0.1                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose : Initializes and handles the hardware for embOS.
*/

#define _GNU_SOURCE
#include <stdlib.h>
#include <time.h>
#include <signal.h>
#include <unistd.h>

#include "RTOS.h"
#include "UDPCOM.h"
#if ((OS_SUPPORT_TRACE_API != 0) && !(OS_64BIT))
#include "SEGGER_SYSVIEW_Sim.h"
#endif

/*********************************************************************
*
*       Configuration
*
**********************************************************************
*/
//
// Specifies how the system tick is incremented.
//
// MULTIPLE_TICKS_PER_INT != 0: (default)
//   With each system tick interrupt the system tick is 0 to n times depending to the actual time passed since the last system tick.
//   Depending on the workload of the CPU and how long the system tick ISR was delayed it is possible that the system tick is incremented multiple times by a single system tick interrupt.
// MULTIPLE_TICKS_PER_INT == 0:
//   With each system tick interrupt the system tick is incremented only once no matter how much time passed since the last system tick interrupt.
//   This results in the embOS time not being equal to the actual time that passed.
//
#ifndef   MULTIPLE_TICKS_PER_INT
  #define MULTIPLE_TICKS_PER_INT  (1)
#endif

/*********************************************************************
*
*       System tick settings
*/
#define OS_TIMER_FREQ       (1000000)  // We're using microseconds for cycles. Thus the frequency of the us counter is 1000000 Hz.
#define OS_TICK_FREQ        (1000)
#define OS_INT_FREQ         (OS_TICK_FREQ)
#define OS_CYCLES_PER_TICK  (OS_TIMER_FREQ / OS_TICK_FREQ)

/*********************************************************************
*
*       embOSView settings
*/
#ifndef   OS_VIEW_IFSELECT
#if (OS_64BIT == 1)
  #define OS_VIEW_IFSELECT  OS_VIEW_DISABLED  // embOSView communication currently not supported for devices with 64-bit address range
#else
  #define OS_VIEW_IFSELECT  OS_VIEW_IF_ETHERNET
#endif
#endif

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/
static OS_U64 _Timestamp;

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _GetTimestamp()
*
*  Function description
*    Calculates and returns the current timestamp in microseconds.
*/
static OS_U64 _GetTimestamp(void) {
  struct timespec Timespec;
  OS_U64          Timestamp;

  clock_gettime(CLOCK_MONOTONIC, &Timespec);
  Timestamp  = Timespec.tv_sec * 1000000;
  Timestamp += Timespec.tv_nsec / 1000;
  return Timestamp;
}

/*********************************************************************
*
*       _ISRTickThread()
*
*  Function description
*    This function simulates the system tick ISR.
*/
static void _ISRTickThread(void) {
  sigset_t   SigSet;
  int        Diff;
  OS_U64     Timestamp;
  useconds_t Duration;
#if (MULTIPLE_TICKS_PER_INT != 0)
  div_t      Ticks;
#endif

#if ((OS_SUPPORT_TRACE_API != 0) && !(OS_64BIT))
  SEGGER_SYSVIEW_X_SetISRName("Tick ISR");
#endif
  //
  // Initialize the variable containing the time to spent until the next tick occurs.
  // It is dynamically adjusted, because with the signal implementation usleep() will return each time this thread receives a signal.
  //
  Duration = OS_CYCLES_PER_TICK;
  //
  // Block all signals for this thread.
  //
  sigemptyset(&SigSet);
  sigfillset(&SigSet);
  sigdelset(&SigSet, SIGUSR1);
  pthread_sigmask(SIG_BLOCK, &SigSet, NULL);
  //
  // Get an initial timestamp.
  //
  _Timestamp = _GetTimestamp();
  //
  // Loop simulating the system tick interrupt.
  //
  while (1) {
    usleep(Duration);
    OS_INT_Enter();
    Timestamp = _GetTimestamp();
    Diff      = (int)((OS_I64)Timestamp - (OS_I64)_Timestamp);
    if (Diff >= OS_CYCLES_PER_TICK) {
#if (MULTIPLE_TICKS_PER_INT != 0)
      Ticks = div(Diff, OS_CYCLES_PER_TICK);
      while (Ticks.quot-- > 0) {
        OS_TICK_Handle();
      }
      _Timestamp = Timestamp - Ticks.rem;         // Remove remaining cycles that belong to the next tick and store the new timestamp.
      Duration = OS_CYCLES_PER_TICK - Ticks.rem;  // Subtract the time that already passed from the tick period.
#else
      OS_TICK_Handle();
      _Timestamp = Timestamp;                     // Remove remaining cycles that belong to the next tick and store the new timestamp.
      Duration = OS_CYCLES_PER_TICK;              // Subtract the time that already passed from the tick period.
#endif
    } else {
      //
      // If usleep() returned earlier than expected, subtract the time already passed from the tick period.
      // The next call of to usleep() will then wait only for the remaining time until the next system tick has to be handled.
      //
      Duration = OS_CYCLES_PER_TICK - Diff;
    }
    OS_INT_Leave();
  }
}

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
*  Function description
*    Returns the current hardware timer count value.
*
*  Return value
*    Current timer count value.
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  unsigned int Cycles;

  Cycles = _GetTimestamp() - _Timestamp;
#if (MULTIPLE_TICKS_PER_INT == 0)
  //
  // Make sure that the calculated cycles that passed since the last tick don't exceed one tick.
  // Else, this could result in OS_TIME_Get_Cycles() returning less cycles than returned by a previous call to OS_TIME_Get_Cycles().
  // This is ensured by capping the returned cycles at OS_CYCLES_PER_TICK - 1.
  //
  if (Cycles > OS_CYCLES_PER_TICK - 1) {
    Cycles = OS_CYCLES_PER_TICK - 1;
  }
#endif
  return Cycles;
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
*  Function description
*    Returns if the hardware timer interrupt pending flag is set.
*
*  Return value
*    == 0: Interrupt pending flag not set.
*    != 0: Interrupt pending flag set.
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  return 0;
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialize the hardware required for embOS to run.
*/
void OS_InitHW(void) {
  OS_INT_IncDI();
  OS_SIM_CreateISRThreadEx(_ISRTickThread, "System Tick ISR");
  {
    OS_SYSTIMER_CONFIG SysTimerConfig = {OS_TIMER_FREQ, OS_INT_FREQ, OS_TIMER_UPCOUNTING, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};
    OS_TIME_ConfigSysTimer(&SysTimerConfig);
  }
  //
  // Initialize communication for embOSView
  //
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Init();
#endif
  //
  // Configure and initialize SEGGER SystemView
  //
#if ((OS_SUPPORT_TRACE_API != 0) && !(OS_64BIT))
  SEGGER_SYSVIEW_Conf();
#endif
  OS_INT_DecRI();
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character.
*/
void OS_COM_Send1(OS_U8 c) {
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Send1(c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)
  OS_USE_PARA(c);          // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
#endif
}

/*************************** End of file ****************************/
